{-# LANGUAGE OverloadedStrings #-}
import qualified Data.List.NonEmpty as NE
import qualified Data.Map as Map
import qualified Data.Set as Set
import qualified Data.Text as T
import qualified Data.Text.IO as TIO

import Control.Applicative ((<|>),pure)
import Control.Arrow (first,second)
import Control.Exception (tryJust)
import Control.Monad (guard,when)
import Data.Char (isAlphaNum,isSpace,toLower)
import Data.Either (partitionEithers)
import Data.Function (on)
import Data.List (find,isPrefixOf)
import Data.Maybe (mapMaybe)
import Data.Semigroup ((<>))
import LDAP (LDAP,LDAPEntry(LDAPEntry),LDAPScope(LdapScopeBase,LdapScopeSubtree),SearchAttributes(LDAPAttrList),
             ldapExternalSaslBind,ldapInitialize,ldapSearch,ldapSimpleBind)
import Network.BSD (getHostByName,getHostName,hostName,hostAliases)
import System.Directory (getDirectoryContents,removeFile,renameFile)
import System.FilePath ((</>))
import System.IO.Error (isDoesNotExistError)

import PidLock (withLockedPidFile)

localDir, targetDir, tmpName :: String
localDir  = "/etc/ssh/authorized_keys"
targetDir = "/var/lib/ssh/authorized_keys"
tmpName   = "/var/lib/ssh/keys.tmp"

tError :: [T.Text] -> a
tError = error . T.unpack . T.unwords

-- return a singleton list value for easier gathering
startWord :: T.Text -> Maybe (T.Text, NE.NonEmpty T.Text)
startWord t
  | not (T.null start) && T.head start /= '#' = Just (start,pure $ T.strip rest)
  | otherwise = Nothing
  where (start,rest) = T.break isSpace $ T.dropWhile isSpace t

getTextOpts :: Map.Map T.Text (NE.NonEmpty a) -> T.Text -> NE.NonEmpty a
getTextOpts conf name = Map.findWithDefault (tError ["No",name,"option found"]) name conf

getTextOpt :: Map.Map T.Text (NE.NonEmpty a) -> T.Text -> a
getTextOpt conf name = case NE.uncons (getTextOpts conf name) of
                         (v,Nothing) -> v
                         _ -> tError ["Multiple",name,"options found"]

-- label with DN for later pruning of the member list
extractKeys :: LDAPEntry -> (String,(String, Set.Set T.Text))
extractKeys (LDAPEntry dn as) =
  case Map.lookup "uid" attrs of
    Nothing -> error $ unwords ["No uid for dn",dn]
    Just [uid] -> (dn,(uid,Set.fromList $ map T.pack $ concatMap findAttr keyAttrs))
    Just _ -> error $ unwords ["Multiple uids for dn",dn] -- FIXME support this somehow
  where attrs = Map.fromList $ map (first (map toLower)) as
        findAttr a = Map.findWithDefault [] a attrs

classify :: LDAPEntry -> Either [String] (String,(String, Set.Set T.Text)) -- either list of member DNs or a user
classify (LDAPEntry _ [(a,ms)]) | map toLower a `elem` memberAttrs = Left ms -- group; after gathering all users prune and resolve
classify user = Right $ extractKeys user -- must be a user

tryJustNotExist :: IO a -> IO (Either () a)
tryJustNotExist = tryJust (guard . isDoesNotExistError)

nilIfNotExist :: (b -> [a]) -> IO b -> IO [a]
nilIfNotExist transform = fmap (either (const []) transform) . tryJustNotExist

readLines :: FilePath -> IO [T.Text]
readLines file = T.lines `nilIfNotExist` TIO.readFile file

fetchKeys :: LDAP -> String -> IO (String, Set.Set T.Text)
fetchKeys ldap dn = do
  entry <- ldapSearch ldap (Just dn) LdapScopeBase Nothing (LDAPAttrList userAttrs) False
  let (_,user) = case entry of
        [] -> error $ unwords ["Search returned no hits for",dn]
        [e] -> extractKeys e
        _ -> error $ unwords ["Search returned multiple hits for",dn]
  return user

refreshKeys :: (FilePath, Set.Set T.Text) -> IO ()
refreshKeys (uid,keys)
  | Set.null keys = putStrLn (unwords ["No keys for",uid]) >> verboseDel uid
  | otherwise = do
  currentKeys <- fmap Set.fromList $ readLines targetName
  when (keys /= currentKeys) $ do
    TIO.writeFile tmpName (T.unlines $ Set.toList keys)
    renameFile tmpName targetName
    putStrLn $ unwords ["Refreshed keys for",uid]
  where targetName = targetDir </> uid

verboseDel :: FilePath -> IO ()
verboseDel uid = do
  res <- tryJustNotExist $ removeFile $ targetDir </> uid
  case res of
    Left ()  -> putStrLn $ unwords ["Keys for",uid,"already disappeared"]
    Right () -> putStrLn $ unwords ["Deleted keys for",uid]

dirEntries :: FilePath -> IO [String]
dirEntries dir = filter (`notElem` [".",".."]) `nilIfNotExist` getDirectoryContents dir

readKeys :: FilePath -> IO (FilePath, Set.Set T.Text)
readKeys uid = do
  keys <- readLines $ localDir </> uid
  return (uid,Set.fromList keys)

-- Lower case only and FIXME handle case insensitivity better
keyAttrs, memberAttrs, userAttrs :: [String]
keyAttrs = ["usercertificate;ssh","sshpublickey"]
memberAttrs = ["member","uniquemember"]
userAttrs = "uid":keyAttrs

main :: IO ()
main = withLockedPidFile "/var/run/ssh-ldap-keys.pid" $ do
  conf <- fmap (Map.fromListWith (flip (<>)) . mapMaybe startWord . T.lines) $ TIO.readFile "/etc/nslcd.conf"
  hostname <- getHostName
  hostEntry <- getHostByName hostname
  let [base,binddn,bindpw] = map (T.unpack . getTextOpt conf) ["base","binddn","bindpw"]
      uri = T.unpack $ T.unwords $ NE.toList $ getTextOpts conf "uri"
      allUsersFilter = T.unpack $ T.concat $ start : map (T.concat . substitute . T.span isAlphaNum) dollars
        where start:dollars = T.split (== '$') (NE.head $ getTextOpts conf "pam_authz_search") -- FIXME handle escaping
              substitute (varName,rest) = [Map.findWithDefault (T.singleton '*') varName replacements,rest]
                where replacements = Map.fromList [("service","sshd"),("hostname",T.pack hostname),("fqdn",T.pack fqdn)]
                      Just fqdn = find ((isPrefixOf `on` map toLower) (hostname ++ ".")) (hostName hostEntry:hostAliases hostEntry) <|>
                                  find (elem '.') (hostName hostEntry:hostAliases hostEntry) <|> Just hostname
  ldap <- ldapInitialize uri
  case fmap NE.uncons (Map.lookup "sasl_mech" conf) of
    Nothing -> ldapSimpleBind ldap binddn bindpw
    Just (mech,Nothing) | T.toLower mech == "external" -> ldapExternalSaslBind ldap
                        | otherwise -> tError ["Unsupported SASL mechanism:",mech]
    _ -> error "Multiple SASL mechanisms specified"

  -- FIXME warn on search size/time limit exceeded conditions
  (memberDNs,(directKeyDNs,directKeys)) <- fmap (second unzip . partitionEithers . map classify) $
    ldapSearch ldap (Just base) LdapScopeSubtree (Just allUsersFilter) (LDAPAttrList $ userAttrs ++ memberAttrs) False
  let unknownMembers = Set.fromList (concat memberDNs) `Set.difference` Set.fromList directKeyDNs
  ldapKeys  <- fmap Map.fromList $ mapM (fetchKeys ldap) $ Set.toList unknownMembers
  localKeys <- fmap Map.fromList $ mapM readKeys =<< dirEntries localDir
  let allKeys = Map.unionsWith Set.union [Map.fromList directKeys,ldapKeys,localKeys]
  presentKeys <- fmap Set.fromList $ dirEntries targetDir
  mapM_ verboseDel $ Set.toList $ presentKeys `Set.difference` (Map.keysSet allKeys)
  mapM_ refreshKeys $ Map.toList allKeys
